from .model import Model
from .drawing import draw
from .fetch import fetch_models, fetch_model, models_kv, save_model, images_kv
from .extract import extract_boxes, Pads

__all__ = [
  'Model', 'draw', 'extract_boxes', 'Pads', 'images_kv',
  'fetch_models', 'fetch_model', 'models_kv', 'save_model'
]


from typing import Sequence
from functools import cached_property
from pydantic import BaseModel
import numpy as np
from .defs import Vec2, block_cols, box_positions, column_offsets

class Model(BaseModel):
  boxWidth: float
  """Box Width relative to 1"""
  rows: int
  """Number of rows"""
  columns: Sequence[float|None]
  """Columns offsets (relative to 1), where column block are represented by `None` (they have width `2*boxWidth`)"""

  @cached_property
  def box_size(self) -> Vec2:
    """Box size normalized to the grid being `(1, 1)`"""
    return self.boxWidth, 1/self.rows
  
  @property
  def block_cols(self) -> Sequence[float]:
    """x-positions of block cols (relative to the grid width being 1)"""
    return block_cols(self.columns, self.boxWidth)

  @cached_property
  def box_positions(self) -> Sequence[Vec2]:
    return box_positions(self.block_cols, self.rows, self.boxWidth)
  
  @cached_property
  def col_offsets(self) -> Sequence[float]:
    return column_offsets(self.columns, self.boxWidth)
  
  @cached_property
  def col_positions(self) -> Sequence[float]:
    return np.cumsum([0, *self.col_offsets])
  
  @cached_property
  def row_positions(self) -> Sequence[float]:
    return [i/self.rows for i in range(self.rows+1)]
  
  @property
  def nboxes(self) -> int:
    return len(self.box_positions)
  
  @property
  def rmin(self):
    return 1 / self.rows
  
  @property
  def cmin(self):
    return min(self.col_offsets)


import numpy as np
import pure_cv as vc
from .model import Model

def draw(
  img: vc.Img, model: Model, *,
  s: np.ndarray, t: np.ndarray,
  thickness: int = 3, color: int | tuple[int, int, int] = (255, 0, 0)
):
  """Draws the `model`'s grid onto `img`
  - `s`: scale
  - `t`: translation
  """
  h, w = img.shape[:2]
  s = s * [w, h]
  t = t * [w, h]
  cols = s[0]*np.array(model.col_positions) + t[0]
  rows = s[1]*np.array(model.row_positions) + t[1]
  out = vc.draw.vlines(img, cols, thickness=thickness, color=color)
  return vc.draw.hlines(out, rows, thickness=thickness, color=color)


from typing import Literal
from ..model import Model
from .andorra import ANDORRA

ModelID = Literal['fcde', 'llobregat23', 'andorra']
MODEL_IDS: list[ModelID] = ['fcde', 'llobregat23', 'andorra']
models: dict[ModelID, Model] = {
  'fcde': Model(
    boxWidth=0.14509394397496592,
    rows=25,
    columns=[None, 0.02818371763906058, 0.03653445043604164, None, 0.02818371763906058, 0.03653445043604164, None]
  ),
  'llobregat23': Model(
    boxWidth=0.23350253819201386,
    rows=30,
    columns=[None, 0.06598984723194462, None]
  ),
  'andorra': ANDORRA
}



from ..import Model

SPACE_W = 0.01904761904761905
NUM_W = 0.05238095238095238

ANDORRA = Model(
  boxWidth=0.14285714285714285,
  rows=20,
  columns=[None, SPACE_W, NUM_W, None, SPACE_W, NUM_W, None]
)


from typing import TypedDict, Annotated, NotRequired
import numpy as np
from .model import Model

class RequiredPads(TypedDict):
  l: float; r: float; t: float; b: float

class Pads(TypedDict):
  l: NotRequired[float]; r: NotRequired[float]; t: NotRequired[float]; b: NotRequired[float]

def unpack(*, l: float, r: float, t: float, b: float):
  return l, r, t, b

Vec2 = tuple[float, float] | np.ndarray

default_pads = RequiredPads(l=0.1, r=0.1, t=0.2, b=0.2)

def padded_rois(
  top_lefts: Annotated[np.ndarray, 'N 2'],
  size: Vec2, *, pads: Pads = {}
) -> Annotated[np.ndarray, 'N 4']:
  """Returns `l, r, t, b` coords of ROIs
  - `top_lefts`: top-left positions of each box (shape `N x 2`)
  - `size`: ROI size
  - `pads`: relative paddings (relative to `size`)
  """
  l, r, t, b = unpack(**{ **default_pads,  **pads })
  tl = top_lefts
  rois = np.zeros((tl.shape[0], 4))
  rois[:, 0] = tl[:, 0] - l * size[0]
  rois[:, 1] = tl[:, 0] + (1 + r) * size[0]
  rois[:, 2] = tl[:, 1] - t * size[1]
  rois[:, 3] = tl[:, 1] + (1 + b) * size[1]
  return np.round(rois).astype(int).clip(0)

def extract_boxes(model: Model, img: np.ndarray, *, tl: Vec2, size: Vec2, pads: Pads = {}) -> list[np.ndarray]:
  h, w = img.shape[:2]
  box_positions = (np.array(model.box_positions)*size + tl) * [w, h]
  box_size = np.array(model.box_size) * [w, h] * size
  rois = padded_rois(box_positions, box_size, pads=pads)
  return [img[t:b, l:r] for l, r, t, b in rois]


Vec2 = tuple[float, float]

def column_offsets(cols: list[float|None], box_width: float) -> list[float]:
  offsets = []
  for c in cols:
    offsets.extend([box_width, box_width] if c is None else [c])
  return offsets

def num_blocks(cols: list[float|None]) -> int:
  return len([x for x in cols if x is None])

def block_cols(cols: list[float|None], box_width: float) -> list[float]:
  """x-positions of block cols (relative to the grid width being 1)"""
  xs = []
  dx = 0
  for c in cols:
    if c is None:
      xs.append(dx)
      dx += 2*box_width
    else:
      dx += c
  return xs

def box_positions(block_cols: list[float], rows: int, box_width: float) -> list[Vec2]:
  ps: list[Vec2] = []
  for c in block_cols:
    for i in range(rows):
      r = i/rows
      ps.append([c, r])
      ps.append([c+box_width, r])
  return ps

def box_size(rows: int, box_width: float) -> Vec2:
    """Box size normalized to the grid being `(1, 1)`"""
    return box_width, 1/rows


from typing import Literal
from dataclasses import dataclass
from haskellian import Either, Left, Right
from pydantic import ValidationError
from azure.storage.blob.aio import ContainerClient, BlobServiceClient
from azure.core.exceptions import ResourceNotFoundError
from .model import Model

STORAGE_ACCOUNT = 'https://movereadcdn.blob.core.windows.net/'
CONTAINER = 'scoresheet-models'

@dataclass
class NotFound:
  tag: Literal['not-found']
  detail: str

@dataclass
class InvalidData:
  tag: Literal['invalid-data']
  detail: str

@dataclass
class UnknownErr:
  tag: Literal['unknown']
  exception: Exception

FetchErr = NotFound | InvalidData | UnknownErr

async def fetch_models() -> Either[NotFound | UnknownErr, list[str]]:
  async with ContainerClient(STORAGE_ACCOUNT, CONTAINER) as cc:
    try:
      return Right([name.split('.')[0] async for name in cc.list_blob_names()])
    except ResourceNotFoundError as e:
      return Left(NotFound('not-found', str(e)))
    except Exception as e:
      return Left(UnknownErr('unknown', e))

async def fetch_model(id: str) -> Either[FetchErr, Model]:
  async with ContainerClient(STORAGE_ACCOUNT, CONTAINER) as cc:
    try:
      blob = await cc.download_blob(id + '.json')
      data = await blob.readall()
      return Right(Model.model_validate_json(data))
    except ResourceNotFoundError as e:
      return Left(NotFound('not-found', str(e)))
    except ValidationError as e:
      return Left(InvalidData('invalid-data', str(e)))
    except Exception as e:
      return Left(UnknownErr('unknown', e))
  
def models_kv(conn_str: str):
  from kv.azure.blob import BlobContainerKV
  return BlobContainerKV[Model].validated(
    Model, client=BlobServiceClient.from_connection_string(conn_str),
    container='scoresheet-models',
  )

async def save_model(id: str, model: Model, *, conn_str: str):
  kv = models_kv(conn_str)
  r = await kv.insert(id + '.json', model)
  return r


def images_kv(conn_str: str):
  from kv.azure.blob import BlobContainerKV
  return BlobContainerKV.from_conn_str(conn_str, 'model-images')